/*
 * clock for root window.
 */

/*
 * - finish alarms set up
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>
#include <locale.h>
#include <string.h>
#include <strings.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include "config.h"
#include "options.h"
#include "rootapp.h"

#define EVENTMASK VisibilityChangeMask | ExposureMask

#define ROUND 0
#define RECT  1

/* Global varriables */
static int end_program = FALSE;

static char      *dpyname    = NULL;
static char      *geometry   = NULL;
static char      *lcolorname = NULL;
static char      *ucolorname = NULL;
static int        draw = ROUND;
static int        h, m, s;
static int        lit = 0, unlit = 0;
static int        radius = 5;


/* Command line options and config file entries */
static Option opttab[] = {
    {'h', NULL,      NULL, O_HELP,   NULL,        FALSE, "display help"},
    {'v', NULL,      NULL, O_VERS,   NULL,        FALSE, "display version"},
    {'d', "display", NULL, C_STRING, &dpyname,    FALSE, "display name"},
    {'g', "geom",    NULL, C_STRING, &geometry,   FALSE, "set geometry"},
    {'c', "color",   NULL, C_STRING, &lcolorname, FALSE, "lit color"},
    {'b', "color",   NULL, C_STRING, &ucolorname, FALSE, "unlit color"},
    {'m', NULL,      NULL, C_BOOL,   &draw,       FALSE, "draw rectangles"},
    {'r', "radius",  NULL, C_INT,    &radius,     FALSE, "leds radius / width"},
    {0,   NULL,      NULL, C_NONE,   NULL,        FALSE, NULL}
};



/****** UPDATING FUNCTIONS ******/

void update_now() {
    time_t     tnow;
    struct tm *now;

    time(&tnow);
    now = localtime(&tnow);
    h = now->tm_hour;
    m = now->tm_min;
    s = now->tm_sec;
}


void get_bits(int t, int b[]) {
    int v = t, i;

    for (i = 0 ; i < 4 ; i++) b[i] = 0;
    i = 0;
    while (v != 0) {
        if (i == 4) {
            fprintf(stderr, "Error : number is too big for 4 bits (%d)\n", t);
            exit(1);
        }
        b[i] = v % 2;
        v = v / 2;
        i++;
    }
}


void draw_round_bits(RootApp *app, int t, int r, int p) {
    int i = 0;
    int b[4];

    get_bits(t, b);
    for (i = 0 ; i < 4 ; i++) {
        int x = p;
        int y = 3 * (r + r / 2) - i * (r + r / 2);
        if (b[i]) {
            set_color(lit);
            XDrawArc(display, app->bgpixmap, gc, x, y, r, r, 0, 23040);
            XFillArc(display, app->bgpixmap, gc, x, y, r, r, 0, 23040);
        } else
            set_color(unlit);
            XDrawArc(display, app->bgpixmap, gc, x, y, r, r, 0, 23040);
    }
}


void draw_rect_bits(RootApp *app, int t, int w, int p) {
    int i = 0;
    int b[4];

    get_bits(t, b);
    for (i = 0 ; i < 4 ; i++) {
        int x = p;
        int y = 3 * w - i * w;
        if (b[i]) {
            set_color(lit);
            XDrawRectangle(display, app->bgpixmap, gc, x, y, w, w / 2);
            XFillRectangle(display, app->bgpixmap, gc, x, y, w, w / 2);
        } else
            set_color(unlit);
            XDrawRectangle(display, app->bgpixmap, gc, x, y, w, w / 2);
    }
}


void draw_time(RootApp *app) {
    switch (draw) {
        case ROUND:
            draw_round_bits(app, h / 10, radius, 0);
            draw_round_bits(app, h % 10, radius, radius + radius / 2);

            draw_round_bits(app, m / 10, radius, radius * 4);
            draw_round_bits(app, m % 10, radius, radius * 5 + radius / 2);

            draw_round_bits(app, s / 10, radius, radius * 8);
            draw_round_bits(app, s % 10, radius, radius * 9 + radius / 2);
            break;
        case RECT:
            draw_rect_bits(app, h / 10, radius, 0);
            draw_rect_bits(app, h % 10, radius, radius + radius / 2);

            draw_rect_bits(app, m / 10, radius, radius * 4);
            draw_rect_bits(app, m % 10, radius, radius * 5 + radius / 2);

            draw_rect_bits(app, s / 10, radius, radius * 8);
            draw_rect_bits(app, s % 10, radius, radius * 9 + radius / 2);
            break;
    }
}


/****** PROCESSING FUNCTIONS ******/

void main_loop(RootApp *app) {
    XEvent event;
    int cont = TRUE;

    draw_time(app);
    display_app(app);
    while (cont) {
        nextevent_or_timeout(&event, 1000);
        update_now();
        clear_bg(app);
        draw_time(app);
        display_app(app);
        if (end_program) cont = FALSE;
    }
}


void set_default_geometry(RootApp *app) {
    int x, y, width, height;

    switch (draw) {
        case RECT:
            width = radius * 10 + radius / 2 + 1;
            height = 4 * radius;
            x = 0;
            y = 0;
            break;
        case ROUND:
        default:
            width = radius * 10 + radius / 2 + 1;
            height = 4 * (radius + radius / 2);
            x = 0;
            y = 0;
            break;
    }

    resize(app, geometry, x, y, width, height);
}


/****** SIGNALS ******/

void install_handler();


void signals_handler(int sig) {
    switch (sig) {
        case SIGINT:
        case SIGTERM:
        case SIGQUIT:
            end_program = TRUE;
            break;
    }
}


void install_handler() {
    signal(SIGTERM, signals_handler);
    signal(SIGQUIT, signals_handler);
    signal(SIGINT,  signals_handler);
}


void manage_signals() {
    struct sigaction sa;

    sa.sa_handler = SIG_IGN;
#ifdef SA_NOCLDWAIT
    sa.sa_flags = SA_NOCLDWAIT;
#else
    sa.sa_flags = 0;
#endif
    sigemptyset(&sa.sa_mask);
    sigaction(SIGCHLD, &sa, NULL);
    install_handler();
}


/****** MAIN ******/


int main(int argc, char *argv[]) {
    RootApp    *app;

    manage_signals();
    setlocale(LC_ALL, "");
    parse_argv(opttab, argc, argv);

    if (radius < 3) radius = 3;
    update_now();

    /* X11 init */
    open_x11(dpyname);
    if (lcolorname) lit = get_color(lcolorname);
    if (ucolorname) unlit = get_color(ucolorname);

    /* Creation */
    app = create_app(EVENTMASK, geometry);
    set_default_geometry(app);

    main_loop(app);

    /* Clean */
    close_app(app);
    close_x11();
    free_vars(opttab);

    return 0;
}
